# -*-mode: python; coding: utf-8 -*-
#******************************************************************************\
#* Copyright (c) 2003-2004, Martin Blais
#* All rights reserved.
#*
#* Redistribution and use in source and binary forms, with or without
#* modification, are permitted provided that the following conditions are
#* met:
#*
#* * Redistributions of source code must retain the above copyright
#*   notice, this list of conditions and the following disclaimer.
#*
#* * Redistributions in binary form must reproduce the above copyright
#*   notice, this list of conditions and the following disclaimer in the
#*   documentation and/or other materials provided with the distribution.
#*
#* * Neither the name of the Martin Blais, Furius, nor the names of its
#*   contributors may be used to endorse or promote products derived from
#*   this software without specific prior written permission.
#*
#* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#******************************************************************************\

"""Automatic completion for optparse module.

This module provide automatic bash completion support for programs that use the
optparse module.  The premise is that the optparse options parser specifies
enough information (and more) for us to be able to generate completion strings
esily.  Another advantage of this over traditional completion schemes where the
completion strings are hard-coded in a separate bash source file, is that the
same code that parses the options is used to generate the completions, so the
completions is always up-to-date with the program itself.

In addition, we allow you specify a list of regular expressions or code that
define what kinds of files should be proposed as completions to this file if
needed.  If you want to implement more complex behaviour, you can instead
specify a function, which will be called with the current directory as an
argument.

You need to activate bash completion using the shell script function that comes
with optcomplete (see http://furius.ca/optcomplete for more details).

"""

__version__ = "$Revision: 1.11 $"
__author__ = "Martin Blais <blais@furius.ca>"

## Bash Protocol Description
## -------------------------
##
## `COMP_CWORD'
##      An index into `${COMP_WORDS}' of the word containing the current
##      cursor position.  This variable is available only in shell
##      functions invoked by the programmable completion facilities (*note
##      Programmable Completion::).
##
## `COMP_LINE'
##      The current command line.  This variable is available only in
##      shell functions and external commands invoked by the programmable
##      completion facilities (*note Programmable Completion::).
##
## `COMP_POINT'
##      The index of the current cursor position relative to the beginning
##      of the current command.  If the current cursor position is at the
##      end of the current command, the value of this variable is equal to
##      `${#COMP_LINE}'.  This variable is available only in shell
##      functions and external commands invoked by the programmable
##      completion facilities (*note Programmable Completion::).
##
## `COMP_WORDS'
##      An array variable consisting of the individual words in the
##      current command line.  This variable is available only in shell
##      functions invoked by the programmable completion facilities (*note
##      Programmable Completion::).
##
## `COMPREPLY'
##      An array variable from which Bash reads the possible completions
##      generated by a shell function invoked by the programmable
##      completion facility (*note Programmable Completion::).

#===============================================================================
# EXTERNAL DECLARATIONS
#===============================================================================

import sys, os, os.path
import types
import re

from pprint import pformat
from collections import namedtuple

from optparse import SUPPRESS_HELP

__all__ = [
    "extract_word",
    "get_comp_words"
    "autocomplete",
    "CompletionResult",
    "FileCompleter",
    "EX_COMPL_STANDARD",
    "EX_COMPL_INVALID",
    "EX_COMPL_NOSPACE",
    "EX_COMPL_FILE",
]

#===============================================================================
# PUBLIC DECLARATIONS
#===============================================================================

debugfn = None # for debugging only

#-------------------------------------------------------------------------------
#

# Unix standard exit codes occupy roughly 60-80 (e.g. os.EX_USAGE=64) as
# specified in /usr/bin/sysexits.h. We claim the range 50-59 for our tab
# completion exit codes, that subsequently dictate further completion operations
# in the shell.
EX_COMPL_STANDARD = 50
EX_COMPL_INVALID  = 51
EX_COMPL_NOSPACE  = 52
EX_COMPL_FILE     = 53

class CompletionResult(namedtuple("CompletionResult", ["completions", "nospace"])):
    """Completion functions should return a CompletionResult"""
    def __new__(cls, *, completions=[], nospace=False):
        return super(CompletionResult, cls).__new__(cls, completions, nospace)

class FileCompleter:

    """Completes by listing all files and directories in current directory."""

    def __call__(self, pwd, line, point, prefix, suffix):
        # Defer file completion to the shell
        sys.exit(EX_COMPL_FILE)

#-------------------------------------------------------------------------------
#
class NoneCompleter:

    """Generates empty completion list."""

    def __call__( self, pwd, line, point, prefix, suffix ):
        return CompletionResult(completions=[], nospace=False)

#-------------------------------------------------------------------------------
#
class RegexCompleter:

    """Completes by filtering all possible files with the given list of
    regexps."""

    def __init__( self, regexlist ):
        if isinstance(regexlist, str):
            regexlist = [regexlist]
        self.regexlist = []
        for r in regexlist:
            if isinstance(r, str):
                r = re.compile(r)
            self.regexlist.append(r)

    def __call__( self, pwd, line, point, prefix, suffix ):
        files = os.listdir(pwd)
        ofiles = []
        for fn in files:
            for r in self.regexlist:
                if r.match(fn):
                    ofiles.append(fn)
                    break
        return CompletionResult(completions=ofiles, nospace=False)

#-------------------------------------------------------------------------------
#
class ListCompleter:

    """Completes by filtering using a fixed list of strings."""

    def __init__( self, stringlist ):
        self.olist = stringlist

    def __call__( self, pwd, line, point, prefix, suffix ):
        return CompletionResult(completions=self.olist, nospace=False)

#-------------------------------------------------------------------------------
#
def extract_word( line, point ):

    """Return a prefix and suffix of the enclosing word.  The character under
    the cursor is the first character of the suffix."""

    wsre = re.compile('[ \t]')

    if point < 0 or point > len(line):
        return '', ''

    preii = point - 1
    while preii >= 0:
        if wsre.match(line[preii]):
            break
        preii -= 1
    preii += 1

    sufii = point
    while sufii < len(line):
        if wsre.match(line[sufii]):
            break
        sufii += 1

    return line[preii : point], line[point : sufii]

#-------------------------------------------------------------------------------
#
def get_comp_words():
    """Get environment variables COMP_WORD_n."""
    cwords = []
    for cw in ('COMP_WORD_%d' % i \
               for i in range(int(os.environ['N_COMP_WORDS']))):
        if os.environ[cw] != '':
            cwords.append(os.environ[cw])
    return cwords

#-------------------------------------------------------------------------------
#
def autocomplete( parser,
                  arg_completer=None, # means use default.
                  opt_completer=None,
                  subcmd_completer=None,
                  subcommands=None,
                  disable_files=False,
                  directories=True,
                  exit_code=EX_COMPL_STANDARD):

    """Automatically detect if we are requested completing and if so generate
    completion automatically from given parser.

    'parser' is the options parser to use.

    'file_completer' is a callable object that gets invoked to produce a list of
    completions for arguments completion (oftentimes files).

    'opt_completer' is the default completer to the options that require a
    value. 'subcmd_completer' is the default completer for the subcommand
    arguments.

    If 'subcommands' is specified, the script expects it to be a map of
    command-name to an object of any kind.  We are assuming that this object is
    a map from command name to a pair of (options parser, completer) for the
    command. If the value is not such a tuple, the method
    'autocomplete(completer)' is invoked on the resulting object.

    This will attempt to match the first non-option argument into a subcommand
    name and if so will use the local parser in the corresponding map entry's
    value.  This is used to implement completion for subcommand syntax and will
    not be needed in most cases."""

    # If we are not requested for complete, simply return silently, let the code
    # caller complete. This is the normal path of execution.
    if 'OPTPARSE_AUTO_COMPLETE' not in os.environ:
        return

    # Set default completers.
    if not arg_completer:
        arg_completer = FileCompleter()
    if not opt_completer:
        opt_completer = FileCompleter()
    if not subcmd_completer:
        subcmd_completer = arg_completer

    # By default, completion will be arguments completion, unless we find out
    # later we're trying to complete for an option.
    completer = arg_completer

    #
    # Completing...
    #

    # FIXME: This is also partially duplicated in optionparse for
    # unclear reasons and should be merged at some point

    # Fetching inputs... not sure if we're going to use these.
    cwords = get_comp_words()
    cline = os.environ['COMP_LINE']
    cpoint = int(os.environ['COMP_POINT'])
    cword = int(os.environ['COMP_CWORD'])


    # If requested, try subcommand syntax to find an options parser for that
    # subcommand.
    if subcommands:
        assert isinstance(subcommands, dict)
        value = guess_first_nonoption(parser, subcommands)
        if value:
            if isinstance(value, list) or \
               isinstance(value, tuple):
                parser = value[0]
                if len(value) > 1 and value[1]:
                    # override completer for command if it is present.
                    completer = value[1]
                else:
                    completer = subcmd_completer
                return autocomplete(parser, completer, exit_code=exit_code)
            else:
                # Call completion method on object. This should call
                # autocomplete() recursively with appropriate arguments.
                if hasattr(value, 'autocomplete'):
                    return value.autocomplete(subcmd_completer, exit_code=exit_code)
                else:
                    sys.exit(exit_code) # no completions for that command object

    # Extract word enclosed word.
    prefix, suffix = extract_word(cline, cpoint)
    # The following would be less exact, but will work nonetheless .
    # prefix, suffix = cwords[cword], None

    # Look at previous word, if it is an option and it requires an argument,
    # check for a local completer.  If there is no completer, what follows
    # directly cannot be another option, so mark to not add those to
    # completions.
    optarg = False
    try:
        # Look for previous word, which will be containing word if the option
        # has an equals sign in it.
        prev = None
        if cword < len(cwords):
            mo = re.search('(--.*)=(.*)', cwords[cword])
            if mo:
                prev, prefix = mo.groups()
        if not prev:
            prev = cwords[cword - 1]

        if prev and prev.startswith('-'):
            option = parser.get_option(prev)
            # The following line change solves the issue when trying to
            # complete more options after one that doesn't take any args.
            if option and option.nargs != None:
                if option.nargs > 0:
                    optarg = True
                    if hasattr(option, 'completer'):
                        completer = option.completer
                    else:
                        completer = opt_completer
                # Warn user at least, it could help him figure out the problem.
                elif hasattr(option, 'completer'):
                    raise SystemExit(
                        "Error: optparse option with a completer "
                        "does not take arguments: %s" % str(option))
    except KeyError:
        pass

    completions = []

    # Either complete with options or results from the assigned completer
    if cword < len(cwords) and cwords[cword].startswith("-"):
        if not optarg and (not prefix or prefix.startswith('-')):
            for option in parser.option_list:
                if not option.help is SUPPRESS_HELP:
                    completions += option._long_opts

    elif completer and (not prefix or not prefix.startswith('-'))\
                   and not disable_files:

        if isinstance(completer, bytes) or \
               isinstance(completer, list) or \
               isinstance(completer, tuple):
            completer = RegexCompleter(completer)
            result = completer(os.getcwd(), cline, cpoint, prefix, suffix)
            completions += result.completions
            if result.nospace:
                exit_code = EX_COMPL_NOSPACE

        elif isinstance(completer, types.FunctionType) or \
             isinstance(completer, types.LambdaType) or \
             isinstance(completer, type) or \
             isinstance(completer, object):
            result = completer(os.getcwd(), cline, cpoint, prefix, suffix)
            completions += result.completions
            if result.nospace:
                exit_code = EX_COMPL_NOSPACE

    # Filter out completions not matching prefix
    if prefix:
        completions = [x for x in completions if x.startswith(prefix)]

    # Print result.
    print(' '.join(completions))

    # Print debug output (if needed).  You can keep a shell with 'tail -f' to
    # the log file to monitor what is happening.
    if debugfn:
        f = open(debugfn, 'a')
        print('---------------------------------------------------------', file=f)
        print('CWORDS', cwords, file=f)
        print('CLINE', cline, file=f)
        print('CPOINT', cpoint, file=f)
        print('CWORD', cword, file=f)
        print('\nShort options', file=f)
        print(pformat(parser._short_opt), file=f)
        print('\nLong options', file=f)
        print(pformat(parser._long_opt), file=f)
        print('Prefix/Suffix:', prefix, suffix, file=f)
        print('completions', completions, file=f)
        f.close()

    # Exit with nonzero exit code (we do not let the caller continue on purpose,
    # this is a run for completions only.)
    sys.exit(exit_code)

#-------------------------------------------------------------------------------
#
def guess_first_nonoption( gparser, subcmds_map ):

    """Given a global options parser, try to guess the first non-option without
    generating an exception. This is used for scripts that implement a
    subcommand syntax, so that we can generate the appropriate completions for
    the subcommand."""

    import copy
    gparser = copy.deepcopy(gparser)
    def print_usage_nousage (self, file=None):
        pass
    gparser.print_usage = print_usage_nousage

    prev_interspersed = gparser.allow_interspersed_args # save state to restore
    gparser.disable_interspersed_args()

    cwords = os.environ['COMP_WORDS'].split()

    try:
        gopts, args = gparser.parse_args(cwords[1:])
    except SystemExit:
        return None

    value = None
    if args:
        subcmdname = args[0]
        try:
            value = subcmds_map[subcmdname]
        except KeyError:
            pass

    gparser.allow_interspersed_args = prev_interspersed # restore state

    return value # can be None, indicates no command chosen.

#-------------------------------------------------------------------------------
#
class CmdComplete:

    """Simple default base class implementation for a subcommand that supports
    command completion.  This class is assuming that there might be a method
    addopts(self, parser) to declare options for this subcommand, and an
    optional completer data member to contain command-specific completion.  Of
    course, you don't really have to use this, but if you do it is convenient to
    have it here."""

    def autocomplete( self, completer ):
        import optparse
        parser = optparse.OptionParser(self.__doc__.strip())
        if hasattr(self, 'addopts'):
            self.addopts(parser)
        if hasattr(self, 'completer'):
            completer = self.completer
        return autocomplete(parser, completer)

#===============================================================================
# TEST
#===============================================================================

def test():
    print(extract_word("extraire un mot d'une phrase", 11))
    print(extract_word("extraire un mot d'une phrase", 12))
    print(extract_word("extraire un mot d'une phrase", 13))
    print(extract_word("extraire un mot d'une phrase", 14))
    print(extract_word("extraire un mot d'une phrase", 0))
    print(extract_word("extraire un mot d'une phrase", 28))
    print(extract_word("extraire un mot d'une phrase", 29))
    print(extract_word("extraire un mot d'une phrase", -2))
    print(extract_word("optcomplete-test do", 19))

if __name__ == '__main__':
    test()
