/*
 * Routines for handling X.509 certificates. Primarily encoding a
 * distinguished name as a string in accordance with RFC 4514.
 *
 * Copyright 2008-2014 Pierre Ossman <ossman@cendio.se> for Cendio AB
 * Copyright 2009-2014 Karl Mikaelsson <derfian@cendio.se> for Cendio AB
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA
 *
 */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>

#include <libtasn1.h>

#include "cert.h"
#include "dn.h"

#include "pkix.asn.h"

struct certificate {
    ASN1_TYPE desc;
    ASN1_TYPE cert;
};

struct crl {
    ASN1_TYPE desc;
    ASN1_TYPE crl;
};

/*
 * Retrieves the DER data for a given extension.
 */
static asn1_retCode find_extension(ASN1_TYPE desc, ASN1_TYPE cert,
                                   const char *extension, char *out,
                                   int *out_len)
{
    asn1_retCode result;

    char val[256];
    int i, val_len, count;
    char namebuf[ASN1_MAX_NAME_SIZE];

    assert(extension);
    assert(out);
    assert(out_len);

    result = asn1_number_of_elements(cert, "tbsCertificate.extensions", &count);
    if (result == ASN1_ELEMENT_NOT_FOUND)
        return ASN1_IDENTIFIER_NOT_FOUND;
    if (result != ASN1_SUCCESS)
        return result;

    for (i = 0; i < count; i++) {
        snprintf(namebuf, sizeof(namebuf),
                 "tbsCertificate.extensions.?%d.extnID", i + 1);

        val_len = sizeof(val);
        result = asn1_read_value(cert, namebuf, val, &val_len);
        if (result != ASN1_SUCCESS)
            return result;

        if (strcmp(val, extension) != 0)
            continue;

        snprintf(namebuf, sizeof(namebuf),
                 "tbsCertificate.extensions.?%d.extnValue", i + 1);

        result = asn1_read_value(cert, namebuf, out, out_len);
        if (result != ASN1_SUCCESS)
            return result;

        return ASN1_SUCCESS;
    }

    return ASN1_IDENTIFIER_NOT_FOUND;
}

/*
 * Reads out the value of a choice.
 */
static asn1_retCode read_choice(ASN1_TYPE cert, const char *source, char *out,
                                size_t maxlen, char *choice_out,
                                size_t choice_maxlen)
{

    asn1_retCode result;
    char choice[64], val[64];
    int val_len, choice_len;
    char namebuf[ASN1_MAX_NAME_SIZE];

    assert(source);
    assert(out);
    assert(maxlen);

    choice_len = sizeof(choice);
    val_len = sizeof(val);

    /* read choice type */
    result = asn1_read_value(cert, source, choice, &choice_len);
    if (result != ASN1_SUCCESS)
        return result;

    if (choice_out) {
        if (choice_len > choice_maxlen)
            return ASN1_MEM_ERROR;

        memcpy(choice_out, choice, choice_len);
    }

    snprintf(namebuf, sizeof(namebuf), "%s.%s", source, choice);

    /* read choice type value */
    result = asn1_read_value(cert, namebuf, val, &val_len);
    if (result != ASN1_SUCCESS)
        return result;

    if (val_len > maxlen)
        return ASN1_MEM_ERROR;

    memcpy(out, val, val_len);

    return ASN1_SUCCESS;
}

/*
 * Get modulus and exponent from RSA public key
 */
static asn1_retCode read_rsa(ASN1_TYPE desc, const char *key, int key_len,
                             char *exp, int *exp_len, char *mod, int *mod_len)
{
    asn1_retCode result;
    ASN1_TYPE pubkey;

    char errstr[ASN1_MAX_ERROR_DESCRIPTION_SIZE];

    assert(key);
    assert(key_len);

    assert(exp);
    assert(exp_len);
    assert(mod);
    assert(mod_len);

    /* Parse the key. */
    result = asn1_create_element(desc, "PKIX1Implicit88.RSAPublicKey", &pubkey);
    if (result != ASN1_SUCCESS)
        return result;

    result = asn1_der_decoding(&pubkey, key, key_len, errstr);
    if (result != ASN1_SUCCESS)
        return result;

    /* Extract data from new element */
    result = asn1_read_value(pubkey, "modulus", mod, mod_len);
    if (result != ASN1_SUCCESS)
        goto free_pubkey;

    result = asn1_read_value(pubkey, "publicExponent", exp, exp_len);
    if (result != ASN1_SUCCESS)
        goto free_pubkey;

    asn1_delete_structure(&pubkey);

    /* Remove padding */
    if ((*mod_len >= 2) && (mod[0] == 0x00) && (mod[1] & 0x80)) {
        *mod_len -= 1;
        memmove(mod, mod + 1, *mod_len);
    }

    if ((*exp_len >= 2) && (exp[0] == 0x00) && (exp[1] & 0x80)) {
        *exp_len -= 1;
        memmove(exp, exp + 1, *exp_len);
    }

    return ASN1_SUCCESS;

  free_pubkey:
    asn1_delete_structure(&pubkey);

    return result;
}

/*
 * Read an algorithm node and return a value from our enum.
 */
static asn1_retCode read_algo(ASN1_TYPE desc, ASN1_TYPE algo,
                              enum pkix_algo *out)
{
    asn1_retCode result;

    char oid[128];
    int oid_len;

    assert(out);

    oid_len = sizeof(oid);
    result = asn1_read_value(algo, "algorithm", &oid, &oid_len);
    if (result != ASN1_SUCCESS)
        return result;

    if (strcmp(oid, "1.2.840.113549.1.1.1") == 0)
        *out = PKIX_ALGO_RSA;
    else if (strcmp(oid, "1.2.840.113549.1.1.2") == 0)
        *out = PKIX_ALGO_RSA_MD2;
    else if (strcmp(oid, "1.2.840.113549.1.1.4") == 0)
        *out = PKIX_ALGO_RSA_MD5;
    else if (strcmp(oid, "1.2.840.113549.1.1.5") == 0)
        *out = PKIX_ALGO_RSA_SHA1;
    else if (strcmp(oid, "1.2.840.113549.1.1.11") == 0)
        *out = PKIX_ALGO_RSA_SHA256;
    else if (strcmp(oid, "1.2.840.113549.1.1.12") == 0)
        *out = PKIX_ALGO_RSA_SHA384;
    else if (strcmp(oid, "1.2.840.113549.1.1.13") == 0)
        *out = PKIX_ALGO_RSA_SHA512;
    else
        return ASN1_IDENTIFIER_NOT_FOUND;

    return ASN1_SUCCESS;
}

/*
 * Convert libtasn1's error codes to POSIX errno:s.
 */
static int mapasn1err(asn1_retCode error)
{
    switch (error) {
    case ASN1_SUCCESS:
        return 0;
    case ASN1_FILE_NOT_FOUND:
        return ENOENT;
    case ASN1_ELEMENT_NOT_FOUND:
        return ENOENT;
    case ASN1_IDENTIFIER_NOT_FOUND:
        return ENOENT;
    case ASN1_DER_ERROR:
        return EINVAL;
    case ASN1_VALUE_NOT_FOUND:
        return ENOENT;
    case ASN1_GENERIC_ERROR:
        return EINVAL;
    case ASN1_VALUE_NOT_VALID:
        return EINVAL;
    case ASN1_TAG_ERROR:
        return EINVAL;
    case ASN1_TAG_IMPLICIT:
        return EINVAL;
    case ASN1_ERROR_TYPE_ANY:
        return EINVAL;
    case ASN1_SYNTAX_ERROR:
        return EINVAL;
    case ASN1_MEM_ERROR:
        return ENOMEM;
    case ASN1_MEM_ALLOC_ERROR:
        return ENOMEM;
    case ASN1_DER_OVERFLOW:
        return ENOMEM;
    case ASN1_NAME_TOO_LONG:
        return ENOMEM;
    case ASN1_ARRAY_ERROR:
        return EINVAL;
    case ASN1_ELEMENT_NOT_EMPTY:
        return EEXIST;
    }

    return EINVAL;
}

/*
 * Reverse the bit order of a bit. Sort of ntohl() but for bits.
 */
static unsigned char invert_byte(unsigned char b)
{
    int i;
    unsigned char result;

    result = 0;
    for (i = 0; i < 8; i++) {
        if (b & (1 << i))
            result |= 1 << (7 - i);
    }

    return result;
}

/*
 * Allocate the structure that holds the parsed certificate.
 */
struct certificate *cert_alloc(void)
{
    struct certificate *cert;

    cert = malloc(sizeof(struct certificate));
    if (!cert)
        return NULL;

    memset(cert, 0, sizeof(struct certificate));

    cert->desc = ASN1_TYPE_EMPTY;
    cert->cert = ASN1_TYPE_EMPTY;

    return cert;
}

/*
 * Free the certificate structure.
 */
void cert_free(struct certificate *cert)
{
    if (!cert)
        return;

    asn1_delete_structure(&cert->cert);
    asn1_delete_structure(&cert->desc);

    free(cert);
}

/*
 * Allocate the structure that holds the parsed revocation list.
 */
struct crl *crl_alloc(void)
{
    struct crl *crl;

    crl = malloc(sizeof(struct crl));
    if (!crl)
        return NULL;

    memset(crl, 0, sizeof(struct crl));

    crl->desc = ASN1_TYPE_EMPTY;
    crl->crl = ASN1_TYPE_EMPTY;

    return crl;
}

/*
 * Free the crl structure.
 */
void crl_free(struct crl *crl)
{
    if (!crl)
        return;

    asn1_delete_structure(&crl->crl);
    asn1_delete_structure(&crl->desc);

    free(crl);
}

static int pkix_desc(ASN1_TYPE * desc)
{
    asn1_retCode result;
    char errstr[ASN1_MAX_ERROR_DESCRIPTION_SIZE];

    if (!desc)
        return -EINVAL;

#ifdef WIN32
    char *tmpbuf;
#else
    char tmpbuf[64];
#endif
    FILE *f;

    /*
     * libtasn1 needs to have the ASN.1 description as a file, so
     * create a temporary file for it.
     */
#ifdef WIN32
    tmpbuf = tempnam(NULL, "pkix");
    if (!tmpbuf)
        return -EIO;

    f = fopen(tmpbuf, "wb");
    if (!f)
        return -EIO;
#else
    int tmpfd;
    strcpy(tmpbuf, "/tmp/pkix.asn.XXXXXX");
    tmpfd = mkstemp(tmpbuf);
    if (tmpfd == -1)
        return -EIO;

    f = fdopen(tmpfd, "wb");
    if (!f) {
        close(tmpfd);
        unlink(tmpbuf);
        return -EIO;
    }
#endif

    if (fputs(pkix_asn, f) < 0) {
        fclose(f);
        unlink(tmpbuf);
        return -EIO;
    }

    fclose(f);

    result = asn1_parser2tree(tmpbuf, desc, errstr);

    unlink(tmpbuf);

    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Parse a certificate in DER form into a libtasn1 object.
 */
int cert_parse(const char *der, int der_len, struct certificate *cert)
{
    int ret;
    asn1_retCode result;
    char errstr[ASN1_MAX_ERROR_DESCRIPTION_SIZE];

    if (!der)
        return -EINVAL;
    if (!cert)
        return -EINVAL;

    ret = pkix_desc(&cert->desc);
    if (ret)
        return ret;

    result = asn1_create_element(cert->desc, "PKIX1Implicit88.Certificate",
                                 &cert->cert);
    if (result != ASN1_SUCCESS)
        goto deldesc;

    result = asn1_der_decoding(&cert->cert, der, der_len, errstr);
    if (result != ASN1_SUCCESS)
        goto delcert;

    return 0;

  delcert:
    asn1_delete_structure(&cert->cert);
  deldesc:
    asn1_delete_structure(&cert->desc);

    return -mapasn1err(result);
}

/*
 * Parse a revocation list in DER form into a libtasn1 object.
 */
int crl_parse(const char *der, int der_len, struct crl *crl)
{
    int ret;
    asn1_retCode result;
    char errstr[ASN1_MAX_ERROR_DESCRIPTION_SIZE];

    if (!der)
        return -EINVAL;
    if (!crl)
        return -EINVAL;

    ret = pkix_desc(&crl->desc);
    if (ret)
        return ret;

    result = asn1_create_element(crl->desc, "PKIX1Implicit88.CertificateList",
                                 &crl->crl);
    if (result != ASN1_SUCCESS)
        goto deldesc;

    result = asn1_der_decoding(&crl->crl, der, der_len, errstr);
    if (result != ASN1_SUCCESS)
        goto delcrl;

    return 0;

  delcrl:
    asn1_delete_structure(&crl->crl);
  deldesc:
    asn1_delete_structure(&crl->desc);

    fprintf(stderr, "Result: %d %x\n", result, result);

    return -mapasn1err(result);
}

/*
 * Construct a string representation of the certificate subject field.
 */
int cert_subject(const struct certificate *cert, char *out, size_t maxlen)
{
    asn1_retCode result;
    ASN1_TYPE dn;

    if (!cert)
        return -EINVAL;
    if (!out)
        return -EINVAL;

    dn = asn1_find_node(cert->cert, "tbsCertificate.subject.rdnSequence");
    if (dn == NULL)
        return -ENOENT;

    result = cert_rfc4514dn(cert->desc, dn, out, maxlen);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Construct a string representation of the certificate subject field.
 */
int cert_subject_dotted_novell(const struct certificate *cert, char *out,
                               size_t maxlen)
{
    asn1_retCode result;
    ASN1_TYPE dn;

    if (!cert)
        return -EINVAL;
    if (!out)
        return -EINVAL;

    dn = asn1_find_node(cert->cert, "tbsCertificate.subject.rdnSequence");
    if (dn == NULL)
        return -ENOENT;

    result = cert_novelldn(cert->desc, dn, out, maxlen);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Fetch a specific attribute from the certificate subject
 * distinguished name.
 */
int cert_subject_attr(const struct certificate *cert, const char *attr,
                      char *out, size_t maxlen)
{
    asn1_retCode result;
    ASN1_TYPE dn;

    if (!cert)
        return -EINVAL;
    if (!out)
        return -EINVAL;

    dn = asn1_find_node(cert->cert, "tbsCertificate.subject.rdnSequence");
    if (dn == NULL)
        return -ENOENT;

    result = cert_dnattr2str(cert->desc, dn, attr, out, maxlen);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Construct a string representation of the certificate issuer field.
 */
int cert_issuer(const struct certificate *cert, char *out, size_t maxlen)
{
    asn1_retCode result;
    ASN1_TYPE dn;

    if (!cert)
        return -EINVAL;
    if (!out)
        return -EINVAL;

    dn = asn1_find_node(cert->cert, "tbsCertificate.issuer.rdnSequence");
    if (dn == NULL)
        return -ENOENT;

    result = cert_rfc4514dn(cert->desc, dn, out, maxlen);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Fetch a specific attribute from the certificate issuer
 * distinguished name.
 */
int cert_issuer_attr(const struct certificate *cert, const char *attr,
                     char *out, size_t maxlen)
{
    asn1_retCode result;
    ASN1_TYPE dn;

    if (!cert)
        return -EINVAL;
    if (!out)
        return -EINVAL;

    dn = asn1_find_node(cert->cert, "tbsCertificate.issuer.rdnSequence");
    if (dn == NULL)
        return -ENOENT;

    result = cert_dnattr2str(cert->desc, dn, attr, out, maxlen);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Construct a string representation of the revocation list issuer field.
 */
int crl_issuer(const struct crl *crl, char *out, size_t maxlen)
{
    asn1_retCode result;
    ASN1_TYPE dn;

    if (!crl)
        return -EINVAL;
    if (!crl)
        return -EINVAL;
    if (!out)
        return -EINVAL;

    dn = asn1_find_node(crl->crl, "tbsCertList.issuer.rdnSequence");
    if (dn == NULL)
        return -ENOENT;

    result = cert_rfc4514dn(crl->desc, dn, out, maxlen);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Read the certificate signature.
 */
int cert_signature(const struct certificate *cert, char *out, size_t * len)
{
    asn1_retCode result;

    if (!cert)
        return -EINVAL;
    if (!out)
        return -EINVAL;
    if (!len)
        return -EINVAL;
    if (!*len)
        return -EINVAL;

    result = asn1_read_value(cert->cert, "signature", out, (int *)len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    /* Length is given as bits */
    *len /= 8;

    return 0;
}

/*
 * Read the certificate signature algorithm.
 */
int cert_signature_algo(const struct certificate *cert, enum pkix_algo *algo)
{
    asn1_retCode result;
    ASN1_TYPE node;

    if (!cert)
        return -EINVAL;
    if (!algo)
        return -EINVAL;

    node = asn1_find_node(cert->cert, "signatureAlgorithm.algorithm");
    if (node == NULL)
        return -ENOENT;

    result = read_algo(cert->cert, node, algo);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Read the revocation list signature.
 */
int crl_signature(const struct crl *crl, char *out, size_t * len)
{
    asn1_retCode result;

    if (!crl)
        return -EINVAL;
    if (!out)
        return -EINVAL;
    if (!len)
        return -EINVAL;
    if (!*len)
        return -EINVAL;

    result = asn1_read_value(crl->crl, "signature", out, (int *)len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    /* Length is given as bits */
    *len /= 8;

    return 0;
}

/*
 * Read the revocation list signature algorithm.
 */
int crl_signature_algo(const struct crl *crl, enum pkix_algo *algo)
{
    asn1_retCode result;
    ASN1_TYPE node;

    if (!crl)
        return -EINVAL;
    if (!algo)
        return -EINVAL;

    node = asn1_find_node(crl->crl, "signatureAlgorithm.algorithm");
    if (node == NULL)
        return -ENOENT;

    result = read_algo(crl->crl, node, algo);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Read the certificate serial.
 */
int cert_serial(const struct certificate *cert, char *out, size_t * len)
{
    asn1_retCode result;

    if (!cert)
        return -EINVAL;
    if (!out)
        return -EINVAL;
    if (!len)
        return -EINVAL;
    if (!*len)
        return -EINVAL;

    result =
        asn1_read_value(cert->cert, "tbsCertificate.serialNumber", out,
                        (int *)len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Read the key usage extension from the certificate. Bits will be
 * returned with the first bit being the low bit, not in the ASN.1
 * BIT STRING hex representation.
 */
int cert_key_usage(const struct certificate *cert, unsigned int *out)
{
    asn1_retCode result;

    char val[256], usage_oid[128];
    int i, val_len;
    char errstr[ASN1_MAX_ERROR_DESCRIPTION_SIZE];

    ASN1_TYPE usage;

    if (!cert)
        return -EINVAL;
    if (!out)
        return -EINVAL;

    val_len = sizeof(usage_oid);
    result = asn1_read_value(cert->desc, "PKIX1Implicit88.id-ce-keyUsage",
                             usage_oid, &val_len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    val_len = sizeof(val);
    result = find_extension(cert->desc, cert->cert, usage_oid, val, &val_len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    usage = ASN1_TYPE_EMPTY;

    result = asn1_create_element(cert->desc, "PKIX1Implicit88.KeyUsage",
                                 &usage);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    result = asn1_der_decoding(&usage, val, val_len, errstr);
    if (result != ASN1_SUCCESS) {
        asn1_delete_structure(&usage);
        return -mapasn1err(result);
    }

    val_len = sizeof(val);
    result = asn1_read_value(usage, "", val, &val_len);
    if (result != ASN1_SUCCESS) {
        asn1_delete_structure(&usage);
        return -mapasn1err(result);
    }

    val_len = (val_len + 7) / 8;

    *out = 0;
    for (i = 0; i < val_len; i++)
        *out |= invert_byte(val[i]) << i * 8;

    asn1_delete_structure(&usage);

    return 0;
}

/*
 * Determines if the certificate represents a Certificate Authority
 */
int cert_is_ca(const struct certificate *cert)
{
    asn1_retCode result;

    char val[256], ext_oid[128];
    int i, val_len;
    char errstr[ASN1_MAX_ERROR_DESCRIPTION_SIZE];
    unsigned int type;

    ASN1_TYPE ext;

    if (!cert)
        return -EINVAL;

    /* First look for the "Basic constraints" extension */
    val_len = sizeof(ext_oid);
    result = asn1_read_value(cert->desc,
                             "PKIX1Implicit88.id-ce-basicConstraints",
                             ext_oid, &val_len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    val_len = sizeof(val);
    result = find_extension(cert->desc, cert->cert, ext_oid, val, &val_len);
    if ((result != ASN1_SUCCESS) && (result != ASN1_IDENTIFIER_NOT_FOUND))
        return -mapasn1err(result);

    if (result == ASN1_SUCCESS) {
        ext = ASN1_TYPE_EMPTY;

        result = asn1_create_element(cert->desc,
                                     "PKIX1Implicit88.BasicConstraints", &ext);
        if (result != ASN1_SUCCESS)
            return -mapasn1err(result);

        result = asn1_der_decoding(&ext, val, val_len, errstr);
        if (result != ASN1_SUCCESS) {
            asn1_delete_structure(&ext);
            return -mapasn1err(result);
        }

        val_len = sizeof(val);
        result = asn1_read_value(ext, "cA", val, &val_len);
        /* libtasn1 cannot handle DEFAULT values */
        if (result == ASN1_ELEMENT_NOT_FOUND)
            strcpy(val, "FALSE");
        else if (result != ASN1_SUCCESS) {
            asn1_delete_structure(&ext);
            return -mapasn1err(result);
        }

        asn1_delete_structure(&ext);

        if (strcmp(val, "TRUE") == 0)
            return 1;
        if (strcmp(val, "FALSE") == 0)
            return 0;

        /* Something has gone horribly wrong */
        return -EINVAL;
    }

    /* Try to find Netscape's "Certificate type" extension */
    val_len = sizeof(ext_oid);
    result = asn1_read_value(cert->desc,
                             "PKIX1Implicit88.id-netscape-certType",
                             ext_oid, &val_len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    val_len = sizeof(val);
    result = find_extension(cert->desc, cert->cert, ext_oid, val, &val_len);
    if ((result != ASN1_SUCCESS) && (result != ASN1_IDENTIFIER_NOT_FOUND))
        return -mapasn1err(result);

    if (result == ASN1_SUCCESS) {
        ext = ASN1_TYPE_EMPTY;

        result = asn1_create_element(cert->desc,
                                     "PKIX1Implicit88.CertType", &ext);
        if (result != ASN1_SUCCESS)
            return -mapasn1err(result);

        result = asn1_der_decoding(&ext, val, val_len, errstr);
        if (result != ASN1_SUCCESS) {
            asn1_delete_structure(&ext);
            return -mapasn1err(result);
        }

        val_len = sizeof(val);
        result = asn1_read_value(ext, "", val, &val_len);
        if (result != ASN1_SUCCESS) {
            asn1_delete_structure(&ext);
            return -mapasn1err(result);
        }

        /* Start by converting the bit string to something sensible */
        val_len = (val_len + 7) / 8;

        type = 0;
        for (i = 0; i < val_len; i++)
            type |= invert_byte(val[i]) << i * 8;

        asn1_delete_structure(&ext);

        /* There are three CA bits in the type field */
        if (type & (1 << 5 | 1 << 6 | 1 << 7))
            return 1;

        return 0;
    }

    /* We couldn't find any fields indicating if this is a CA, so assume
       it isn't. */
    return 0;
}

/*
 * Get modulus and exponent from RSA public key
 */
int cert_rsa_public_key_components(const struct certificate *cert,
                                   char *exp, size_t * exp_len, char *mod,
                                   size_t * mod_len)
{
    asn1_retCode result;
    char rsa_oid[128];
    int oid_len = sizeof(rsa_oid);
    /* Algorithm of public key */
    char algo[128];
    int algolen = sizeof(algo);

    /* Public key */
    char key[8192];
    int keylen = sizeof(key);

    if (!exp || !exp_len || !mod || !mod_len)
        return -EINVAL;

    oid_len = sizeof(rsa_oid);
    result = asn1_read_value(cert->desc, "PKIX1Implicit88.rsaEncryption",
                             rsa_oid, &oid_len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    result = asn1_read_value(cert->cert,
                             "tbsCertificate.subjectPublicKeyInfo.algorithm.algorithm",
                             &algo, &algolen);

    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    /* Check if the algorithm is RSA */
    if (strcmp(algo, rsa_oid) != 0)
        return -EINVAL;

    result = asn1_read_value(cert->cert,
                             "tbsCertificate.subjectPublicKeyInfo.subjectPublicKey",
                             &key, &keylen);

    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    /* Length is in bits and we want bytes */
    keylen /= 8;

    /* Read out the actual key */
    result =
        read_rsa(cert->desc, key, keylen, exp, (int *)exp_len, mod,
                 (int *)mod_len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Get the date from which the certificate is valid
 */
int cert_valid_notbefore(const struct certificate *cert, char *out,
                         size_t maxlen)
{
    asn1_retCode result;

    result =
        read_choice(cert->cert, "tbsCertificate.validity.notBefore", out,
                    maxlen, NULL, 0);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Get the date after which the certificate is invalid
 */
int cert_valid_notafter(const struct certificate *cert, char *out,
                        size_t maxlen)
{
    asn1_retCode result;

    result =
        read_choice(cert->cert, "tbsCertificate.validity.notAfter", out, maxlen,
                    NULL, 0);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Get the date from which the certificate is valid
 */
int crl_update_this(const struct crl *crl, char *out, size_t maxlen)
{
    asn1_retCode result;

    result =
        read_choice(crl->crl, "tbsCertList.thisUpdate", out, maxlen, NULL, 0);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Get the date after which the certificate is invalid
 */
int crl_update_next(const struct crl *crl, char *out, size_t maxlen)
{
    asn1_retCode result;

    result =
        read_choice(crl->crl, "tbsCertList.nextUpdate", out, maxlen, NULL, 0);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}

/*
 * Retrieve the URI for any OCSP services. It's a nice to have, so not
 * implemented yet.
 */
int cert_ocsp(const struct certificate *cert, char *out, size_t maxlen)
{
    return 1;
}

/*
 * Retrieves CRL distribution points from X509 certificate extension
 * data.
 *  cert: certificate in DER form
 *  out: array of char* where results will be stored. caller have to
 *       free the returned char pointers
 *  maxlen: number of elements in out array
 */
int cert_crl(const struct certificate *cert, char *out[], size_t maxlen)
{
    asn1_retCode result;
    char crl_oid[128], val[8192], errstr[ASN1_MAX_ERROR_DESCRIPTION_SIZE];
    char ext[2048], namebuf[ASN1_MAX_NAME_SIZE], *item;
    int len, n, i, ndistpoints, nnames, nresults;
    ASN1_TYPE crl;

    nresults = 0;

    if (!cert || !out)
        return -EINVAL;

    /*
     * The CRL Distribution Points is an extension to X509
     * certificates, so we need to make sure that it's present.
     */

    len = sizeof(crl_oid);
    result = asn1_read_value(cert->desc,
                             "PKIX1Implicit88.id-ce-cRLDistributionPoints",
                             crl_oid, &len);

    if (result != ASN1_SUCCESS) {
        fprintf(stderr, "Failed to read distribution point OID\n");
        asn1_perror(result);
        return -mapasn1err(result);
    }

    len = sizeof(ext);
    result = find_extension(cert->desc, cert->cert, crl_oid, ext, &len);

    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    /*
     * Since the extension is stored as a DER-coded OCTET STRING in
     * the certificate, we need to read it and parse it using libtasn1
     * to be able to access the extension data.
     */

    crl = ASN1_TYPE_EMPTY;

    result = asn1_create_element(cert->desc,
                                 "PKIX1Implicit88.CRLDistPointsSyntax", &crl);

    if (result != ASN1_SUCCESS) {
        fprintf(stderr,
                "Failed to create PKIX1Implicit88.CRLDistPointsSyntax element\n");
        return -mapasn1err(result);
    }

    result = asn1_der_decoding(&crl, &ext, len, errstr);

    if (result != ASN1_SUCCESS) {
        asn1_delete_structure(&crl);
        fprintf(stderr,
                "Failed to decode extension as CRLDistPointsSyntax element!\nError is %s\n",
                errstr);
        asn1_perror(result);
        return -mapasn1err(result);
    }

    /* CRLDistPointsSyntax is made up of a sequence of
       DistributionPoints, which in turn is a sequence of choices and
       sequences.. I'm sure you could encode just about anything with
       this definition. */

    result = asn1_number_of_elements(crl, "", &ndistpoints);

    if (result != ASN1_SUCCESS) {
        fprintf(stderr,
                "Failed to determine the number of distributionPoints in CRLDistPointsSyntax element\n");
        asn1_perror(result);
        asn1_delete_structure(&crl);
        return -mapasn1err(result);
    }

    for (i = 0; i < ndistpoints; i++) {

        /* Try to read the fullName attribute first, and if that
           fails, try the nameRelativeToCRLIssuer */

        snprintf(namebuf, sizeof(namebuf), "?%d.distributionPoint.fullName",
                 i + 1);
        result = asn1_number_of_elements(crl, namebuf, &nnames);

        if (result == ASN1_SUCCESS) {

            /*
             * A fullName structure is a sequence of choices between
             * lots of different string types. So we'll need to handle
             * all sorts of different data types. The most common one
             * with SITHS seems to be an URI, though.
             */

            for (n = 0; n < nnames; n++) {

                snprintf(namebuf, sizeof(namebuf),
                         "?%d.distributionPoint.fullName.?%d", i + 1, n + 1);

                len = sizeof(val);
                memset(val, '\0', len);

                result = asn1_read_value(crl, namebuf, val, &len);

                if (result != ASN1_SUCCESS) {
                    fprintf(stderr, "Error when reading choice type from %s\n",
                            namebuf);
                    asn1_perror(result);
                    asn1_delete_structure(&crl);
                    return -mapasn1err(result);
                }

                /*
                 * We should perform additional checks for the actual
                 * type of value here, in case we need to unpack the
                 * structure even further. In most cases, we're safe
                 * with just reading the data value.
                 */

                snprintf(namebuf, sizeof(namebuf),
                         "?%d.distributionPoint.fullName.?%d.%s", i + 1, n + 1,
                         val);

                len = sizeof(val);
                memset(val, '\0', len);

                result = asn1_read_value(crl, namebuf, val, &len);

                if (result != ASN1_SUCCESS) {
                    fprintf(stderr, "Error when reading choice value from %s\n",
                            namebuf);
                    asn1_perror(result);
                    asn1_delete_structure(&crl);
                    return -mapasn1err(result);
                }


                nresults += 1;

                if (nresults >= maxlen) {
                    return -ENOMEM;
                }

                /*
                 * Allocate a new char pointer that will hold the
                 * result inside the out array and store the result
                 * there.
                 */

                (*out) = (char *)malloc(len + 1);
                item = (*out++);

                if (!item) {
                    perror("cert.c");
                    asn1_delete_structure(&crl);
                    return -ENOMEM;
                }

                memset(item, '\0', len + 1);
                memcpy(item, val, len);
            }

        } else {
            /*
             * FIXME: Read data from the nameRelativeToCRLIssuer here.
             */
            fprintf(stderr,
                    "CRL distribution list not stored in fullName field\n");
            /* ENOSYS: function not implemented */
            return -ENOSYS;
        }
    }

    asn1_delete_structure(&crl);
    return 0;
}

/*
 * Get the number of entries in this revocation list.
 */
int crl_count(const struct crl *crl)
{
    asn1_retCode result;
    int count;

    if (!crl)
        return -EINVAL;

    result =
        asn1_number_of_elements(crl->crl, "tbsCertList.revokedCertificates",
                                &count);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return count;
}

int crl_entry(const struct crl *crl, size_t index, char *out, size_t * len)
{
    asn1_retCode result;
    char namebuf[ASN1_MAX_NAME_SIZE];

    if (!crl)
        return -EINVAL;
    if (!out)
        return -EINVAL;
    if (!len)
        return -EINVAL;
    if (!*len)
        return -EINVAL;

    snprintf(namebuf, sizeof(namebuf),
             "tbsCertList.revokedCertificates.?%u.userCertificate",
             (unsigned)(index + 1));

    result = asn1_read_value(crl->crl, namebuf, out, (int *)len);
    if (result != ASN1_SUCCESS)
        return -mapasn1err(result);

    return 0;
}
