/*
 * Routines for handling X.509 certificates. Primarily encoding a
 * distinguished name as a string in accordance with RFC 4514.
 *
 * Copyright 2008-2014 Pierre Ossman <ossman@cendio.se> for Cendio AB
 * Copyright 2009-2014 Karl Mikaelsson <derfian@cendio.se> for Cendio AB
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA
 *
 */

#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif

    /*
     * Maximum size of the text representation of a RDN.
     *
     * There is no fundamendal upper limit, but most fields are limited
     * to 64 or 128 characters. Worst case scenario is a 128 character
     * UCS-4 string that we need to encode hexadecimally.
     */
#define MAX_RDN_TEXT_SIZE 1024

    /*
     * Maximum size of the text representation of a DN.
     *
     * Again, there is no upper limit, but LDAP referals (which
     * can contain a DN) are limited to 2048 characters. Again,
     * given a UCS-4 encoding with hexadecimal representation gives
     * this size.
     */
#define MAX_DN_TEXT_SIZE 16834

    struct certificate;
    struct crl;

    enum pkix_algo {
        PKIX_ALGO_RSA,
        PKIX_ALGO_RSA_MD2,
        PKIX_ALGO_RSA_MD5,
        PKIX_ALGO_RSA_SHA1,
        PKIX_ALGO_RSA_SHA256,
        PKIX_ALGO_RSA_SHA384,
        PKIX_ALGO_RSA_SHA512,
    };

    struct certificate *cert_alloc(void);
    void cert_free(struct certificate *cert);

    struct crl *crl_alloc(void);
    void crl_free(struct crl *crl);

    int cert_parse(const char *der, int der_len, struct certificate *cert);

    int crl_parse(const char *der, int der_len, struct crl *crl);

    int cert_subject(const struct certificate *cert, char *out, size_t maxlen);
    int cert_subject_attr(const struct certificate *cert, const char *attr,
                          char *out, size_t maxlen);
    int cert_subject_dotted_novell(const struct certificate *cert, char *out,
                                   size_t maxlen);

    int cert_issuer(const struct certificate *cert, char *out, size_t maxlen);
    int cert_issuer_attr(const struct certificate *cert, const char *attr,
                         char *out, size_t maxlen);

    int crl_issuer(const struct crl *crl, char *out, size_t maxlen);

    int cert_signature(const struct certificate *cert, char *out, size_t * len);
    int cert_signature_algo(const struct certificate *cert,
                            enum pkix_algo *algo);

    int crl_signature(const struct crl *crl, char *out, size_t * len);
    int crl_signature_algo(const struct crl *crl, enum pkix_algo *algo);

    int cert_serial(const struct certificate *cert, char *out, size_t * len);

    int cert_key_usage(const struct certificate *cert, unsigned int *out);
    int cert_is_ca(const struct certificate *cert);

    int cert_rsa_public_key_components(const struct certificate *cert,
                                       char *exp, size_t * exp_len, char *mod,
                                       size_t * mod_len);

    int cert_valid_notbefore(const struct certificate *cert, char *out,
                             size_t maxlen);
    int cert_valid_notafter(const struct certificate *cert, char *out,
                            size_t maxlen);

    int crl_update_this(const struct crl *crl, char *out, size_t maxlen);
    int crl_update_next(const struct crl *crl, char *out, size_t maxlen);

    int cert_crl(const struct certificate *cert, char *out[], size_t maxlen);

    int crl_count(const struct crl *crl);
    int crl_entry(const struct crl *crl, size_t index, char *out, size_t * len);

#ifdef __cplusplus
}
#endif
