/*
 * Routines for handling X.509 certificates. Primarily encoding a
 * distinguished name as a string in accordance with RFC 4514.
 *
 * Copyright 2008-2014 Pierre Ossman <ossman@cendio.se> for Cendio AB
 * Copyright 2009-2014 Karl Mikaelsson <derfian@cendio.se> for Cendio AB
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA
 *
 */

#include <assert.h>
#include <string.h>

#include <libtasn1.h>

#include "cert.h"
#include "dn.h"

#include "pkix.asn.h"
#include "oidmap.h"
#include "novell_oidmap.h"

/*
 * Look up an oid and find the LDAP "short name".
 */
static const char *oid_to_name(const char *oid)
{
    int i;

    for (i = 0; i < sizeof(ldap_oid_name) / sizeof(*ldap_oid_name); i++) {
        if (strcmp(oid, ldap_oid_name[i].oid) == 0)
            return ldap_oid_name[i].name;
    }

    return NULL;
}

/*
 * Same as above, but add a few extras defs and overrides for those
 * lovely chaps at Novell, Inc.
 */
static const char *oid_to_novell_name(const char *oid)
{
    int i;

    for (i = 0; i < sizeof(novell_oid_name) / sizeof(*novell_oid_name); i++) {
        if (strcmp(oid, novell_oid_name[i].oid) == 0)
            return novell_oid_name[i].name;
    }

    return oid_to_name(oid);
}

/*
 * Convert a LDAP "short name" to the corresponding oid.
 */
static const char *name_to_oid(const char *name)
{
    int i;

    for (i = 0; i < sizeof(ldap_oid_name) / sizeof(*ldap_oid_name); i++) {
        /*
         * Attribute names are always case insensitive
         *
         * Note: Names are expected to use only ASCII (strictly speaking,
         *       a subset of ASCII) so we should not use locale dependent
         *       case adjustment.
         */
        if (strcasecmp(name, ldap_oid_name[i].name) == 0)
            return ldap_oid_name[i].oid;
    }

    return NULL;
}

/*
 * Make a qualified guess as to if the given blob is an oid or not.
 */
static int is_oid(const char *blob)
{
    if (strspn(blob, "0123456789.") == strlen(blob))
        return 1;

    return 0;
}

/*
 * Compares two oids, where each oid can be a numeric or string
 * representation.
 */
static int oideq(const char *oid1, const char *oid2)
{
    const char *num1, *num2;

    /* Trivial case */
    if (strcmp(oid1, oid2) == 0)
        return 1;

    num1 = num2 = NULL;

    /* Make sure oid 1 is numerical */
    if (is_oid(oid1))
        num1 = oid1;
    else
        num1 = name_to_oid(oid1);

    if (!num1)
        return 0;

    /* Make sure oid 2 is numerical */
    if (is_oid(oid2))
        num2 = oid2;
    else
        num2 = name_to_oid(oid2);

    if (!num2)
        return 0;

    /* Do another check now that both are numerical */
    if (strcmp(num1, num2) == 0)
        return 1;

    return 0;
}


/*
 * Convert a Unicode code point to UTF-8.
 */
static asn1_retCode ucs4_to_utf8(unsigned int wchar, char *out, int *len)
{
    assert(out);
    assert(len);

    if (wchar < 0x80) {
        if (*len < 1)
            return ASN1_MEM_ERROR;

        *out = wchar;

        *len = 1;
    } else if (wchar < 0x800) {
        if (*len < 2)
            return ASN1_MEM_ERROR;

        *out = 0xc0 | (wchar >> 6);
        out++;
        *out = 0x80 | (wchar & 0x3f);

        *len = 2;
    } else if (wchar < 0x10000) {
        if (*len < 3)
            return ASN1_MEM_ERROR;

        *out = 0xe0 | (wchar >> 12);
        out++;
        *out = 0x80 | ((wchar >> 6) & 0x3f);
        out++;
        *out = 0x80 | (wchar & 0x3f);

        *len = 3;
    } else if (wchar < 0x110000) {
        if (*len < 4)
            return ASN1_MEM_ERROR;

        *out = 0xf0 | (wchar >> 18);
        out++;
        *out = 0x80 | ((wchar >> 12) & 0x3f);
        out++;
        *out = 0x80 | ((wchar >> 6) & 0x3f);
        out++;
        *out = 0x80 | (wchar & 0x3f);

        *len = 4;
    } else {
        return ASN1_TAG_ERROR;
    }

    return ASN1_SUCCESS;
}

/*
 * Decode a DirectoryString DER object, converting it to UTF-8
 * in the process.
 */
static asn1_retCode der_to_str(const char *der, int der_len,
                               char *out, size_t maxlen)
{
    asn1_retCode result;
    char err[ASN1_MAX_ERROR_DESCRIPTION_SIZE];
    char type[ASN1_MAX_NAME_SIZE];
    const char *val;
    const unsigned char *pval;
    int val_len, i, len;
    char *orig_out;

    unsigned char cls;
    unsigned long tag;
    int tag_len;

    orig_out = out;

    /*
     * Strings are simple DER data structures so we can just decode
     * them by hand here
     */
    result = asn1_get_tag_der(der, der_len, &cls, &len, &tag);
    if (result != ASN1_SUCCESS)
        return result;

    der += len;
    der_len -= len;

    val_len = asn1_get_length_der(der, der_len, &len);
    if (result < 0)
        return ASN1_TAG_ERROR;

    der += len;
    der_len -= len;

    val = der;

    /*
     * Convert the value to UTF-8.
     */
    if (tag == ASN1_TAG_TELETEX_STRING) {
        /*
         * From RFC2459:
         *
         * In addition, many legacy implementations support names encoded
         * in the ISO 8859-1 character set (Latin1String) but tag them as
         * TeletexString.  The Latin1String includes characters used in
         * Western European countries which are not part of the
         * TeletexString charcter set.  Implementations that process
         * TeletexString SHOULD be prepared to handle the entire ISO 8859-1
         * character set.[ISO 8859-1]
         */
        for (i = 0; i < val_len; i++) {
            len = maxlen;
            result = ucs4_to_utf8((unsigned char)val[i], out, &len);
            if (result != ASN1_SUCCESS)
                return result;

            out += len;
            maxlen -= len;
        }
    } else if (tag == ASN1_TAG_BMP_STRING) {
        /* UCS-2 */
        unsigned short *ucs2;

        if ((val_len % 2) != 0)
            return ASN1_TAG_ERROR;

        ucs2 = (unsigned short *)val;
        for (i = 0; i < val_len / 2; i++) {
            len = maxlen;
            result = ucs4_to_utf8(ucs2[i], out, &len);
            if (result != ASN1_SUCCESS)
                return result;

            out += len;
            maxlen -= len;
        }
    } else if (tag == ASN1_TAG_UNIVERSAL_STRING) {
        /* UCS-4 */
        unsigned int *ucs4;

        assert((val_len % 4) == 0);

        ucs4 = (unsigned int *)val;
        for (i = 0; i < val_len / 4; i++) {
            len = maxlen;
            result = ucs4_to_utf8(ucs4[i], out, &len);
            if (result != ASN1_SUCCESS)
                return result;

            out += len;
            maxlen -= len;
        }
    } else if ((tag == ASN1_TAG_PRINTABLE_STRING) ||
               (tag == ASN1_TAG_IA5_STRING) || (tag == ASN1_TAG_UTF8_STRING)) {
        /*
         * Printable string is a subset of ASCII, and IA5String _is_
         * ASCII, so it's safe to just copy those.
         */
        if (val_len > maxlen)
            return ASN1_MEM_ERROR;

        memcpy(out, val, val_len);
        out += val_len;
        maxlen -= val_len;
    } else {
        // Not a string type
        return ASN1_VALUE_NOT_VALID;
    }

    /*
     * None of the above conversions terminate the string...
     */
    if (maxlen < 1)
        return ASN1_MEM_ERROR;

    *out = '\0';

    /*
     * Embedded \0s will confuse upper layers and is a potential
     * security risk. Return failure if any are detected.
     */
    while (orig_out != out) {
        if (*orig_out == '\0')
            return ASN1_VALUE_NOT_VALID;
        orig_out++;
    }

    result = ASN1_SUCCESS;

    return result;
}

/*
 * Apply RFC 4514 escaping rules.
 */
static asn1_retCode escape(const char *in, char *out, int maxlen)
{
    unsigned char ch;

    assert(in);
    assert(out);

    /*
     * Space and hash aren't allowed as the first character...
     */
    switch (*in) {
    case '#':
    case ' ':
        if (maxlen < 2)
            return ASN1_MEM_ERROR;

        *out = '\\';
        out++;
        *out = *in;
        out++;

        maxlen -= 2;
        break;
    }

    while (*in != '\0') {
        ch = *in;
        if ((ch < 0x20) || (ch > 0x80)) {
            if (maxlen < 4)
                return ASN1_MEM_ERROR;

            sprintf(out, "\\%02x", (unsigned)ch);
            out += 3;
            maxlen -= 3;
        } else {
            switch (ch) {
            case ' ':
                /*
                 * A space at the end needs to be escaped...
                 */
                if (*(in + 1) != '\0') {
                    if (maxlen < 1)
                        return ASN1_MEM_ERROR;

                    *out = ch;
                    out++;
                    maxlen--;
                    break;
                }
                /* Fall through... */
            case '"':
            case '#':
            case '+':
            case ',':
            case ';':
            case '<':
            case '=':
            case '>':
            case '\\':
                if (maxlen < 2)
                    return ASN1_MEM_ERROR;

                *out = '\\';
                out++;
                *out = ch;
                out++;

                maxlen -= 2;
                break;
            default:
                if (maxlen < 1)
                    return ASN1_MEM_ERROR;

                *out = ch;
                out++;
                maxlen--;
            }
        }

        in++;
    }

    if (maxlen < 1)
        return ASN1_MEM_ERROR;

    *out = '\0';

    return ASN1_SUCCESS;
}

static asn1_retCode ava_to_str(ASN1_TYPE desc, ASN1_TYPE dn,
                               int rdnidx, int avaidx,
                               char *out, int maxlen, int novell)
{
    asn1_retCode result;

    char namebuf[ASN1_MAX_NAME_SIZE];
    char val[MAX_RDN_TEXT_SIZE];
    int val_len;
    char oid[MAX_RDN_TEXT_SIZE];
    int oid_len;
    const char *name;

    snprintf(namebuf, sizeof(namebuf),
             "rdnSequence.?%d.?%d.type", rdnidx, avaidx);

    oid_len = sizeof(oid) - 1;
    result = asn1_read_value(dn, namebuf, oid, &oid_len);
    if (result != ASN1_SUCCESS)
        return result;

    if (novell)
        name = oid_to_novell_name(oid);
    else
        name = oid_to_name(oid);

    snprintf(namebuf, sizeof(namebuf),
             "rdnSequence.?%d.?%d.value", rdnidx, avaidx);

    val_len = sizeof(val) - 1;
    result = asn1_read_value(dn, namebuf, val, &val_len);
    if (result != ASN1_SUCCESS)
        return result;

    /*
     * From RFC 4514:
     *
     * If the AttributeType is of the dotted-decimal form, the
     * AttributeValue is represented by an number sign ('#' U+0023)
     * character followed by the hexadecimal encoding of each of the
     * octets of the BER encoding of the X.500 AttributeValue.
     */
    if (name) {
        result = der_to_str(val, val_len, val, sizeof(val));
        if (result == ASN1_SUCCESS) {
            char val2[MAX_RDN_TEXT_SIZE];

            if (maxlen < strlen(name) + 1)
                return ASN1_MEM_ERROR;
            strcpy(out, name);
            out += strlen(name);
            maxlen -= strlen(name);

            if (maxlen < 1)
                return ASN1_MEM_ERROR;
            *out = '=';
            out++;
            maxlen--;

            if (novell)
                strcpy(val2, val);
            else {
                result = escape(val, val2, sizeof(val2));
                if (result != ASN1_SUCCESS)
                    return result;
            }

            if (maxlen < strlen(val2) + 1)
                return ASN1_MEM_ERROR;
            strcpy(out, val2);
            out += strlen(val2);
            maxlen -= strlen(val2);

            return ASN1_SUCCESS;
        } else if (result != ASN1_VALUE_NOT_VALID)
            return result;
    }

    int i;

    if (maxlen < strlen(oid) + 1)
        return ASN1_MEM_ERROR;
    strcpy(out, oid);
    out += strlen(oid);
    maxlen -= strlen(oid);

    if (maxlen < 1)
        return ASN1_MEM_ERROR;
    *out = '=';
    out++;
    maxlen--;

    if (maxlen < 1)
        return ASN1_MEM_ERROR;
    *out = '#';
    out++;
    maxlen--;

    for (i = 0; i < val_len; i++) {
        if (maxlen < 3)
            return ASN1_MEM_ERROR;

        sprintf(out, "%02x", (unsigned)(unsigned char)val[i]);
        out += 2;
        maxlen -= 2;
    }

    if (maxlen < 1)
        return ASN1_MEM_ERROR;
    *out = '\0';
    out++;
    maxlen--;

    return ASN1_SUCCESS;
}

static asn1_retCode rdn_to_str(ASN1_TYPE desc, ASN1_TYPE dn, int rdnidx,
                               char *out, int maxlen, int novell)
{
    asn1_retCode result;

    int i, avas;

    char namebuf[ASN1_MAX_NAME_SIZE];

    /*
     * Iterate over each AVA (Attribute Value Assertion) for
     * this RDN. Normally there is just one.
     */
    snprintf(namebuf, sizeof(namebuf), "rdnSequence.?%d", rdnidx);

    result = asn1_number_of_elements(dn, namebuf, &avas);
    if (result != ASN1_SUCCESS)
        return result;

    for (i = 0; i < avas; i++) {
        int avaidx;
        size_t len;

        /*
         * It is allowed to iterate in either direction here, but we
         * do it backwards in order to be compatible with OpenSSL.
         */
        avaidx = avas - i;

        if (i != 0) {
            if (maxlen < 1)
                return ASN1_MEM_ERROR;
            *out = '+';
            out++;
            maxlen--;
        }

        result = ava_to_str(desc, dn, rdnidx, avaidx, out, maxlen, novell);
        if (result != ASN1_SUCCESS)
            return result;

        len = strlen(out);
        assert(len < maxlen);

        out += len;
        maxlen -= len;
    }

    return ASN1_SUCCESS;
}

/*
 * Create a text representation of a distinguished name in accordance
 * with RFC 4514 or Novell's undocumented syntax
 */
static asn1_retCode dn_to_str(ASN1_TYPE desc, ASN1_TYPE dn, char *out,
                              int maxlen, int novell)
{
    asn1_retCode result;

    int i, rdns;

    assert(out);

    result = asn1_number_of_elements(dn, "rdnSequence", &rdns);
    if (result != ASN1_SUCCESS)
        return result;

    /*
     * Iterate over each RDN (Relative Distinguished Name)
     */
    for (i = 0; i < rdns; i++) {
        int rdnidx;
        size_t len;

        /*
         * Novell iterates forward to back, but RFC 4514 goes backwards
         */
        if (novell)
            rdnidx = i + 1;
        else
            rdnidx = rdns - i;

        if (i != 0) {
            if (maxlen < 1)
                return ASN1_MEM_ERROR;
            if (novell)
                *out = '.';
            else
                *out = ',';
            out++;
            maxlen--;
        }

        result = rdn_to_str(desc, dn, rdnidx, out, maxlen, novell);
        if (result != ASN1_SUCCESS)
            return result;

        len = strlen(out);
        assert(len < maxlen);

        out += len;
        maxlen -= len;
    }

    return ASN1_SUCCESS;
}

/*
 * Create a text representation of a distinguished name in accordance
 * with RFC 4514.
 */
asn1_retCode cert_rfc4514dn(ASN1_TYPE desc, ASN1_TYPE dn, char *out, int maxlen)
{
    return dn_to_str(desc, dn, out, maxlen, 0);
}

/*
 * Create a text representation of a distinguished name in accordance
 * with the odd dotted Novell syntax that iManager manages to produce.
 */
asn1_retCode cert_novelldn(ASN1_TYPE desc, ASN1_TYPE dn, char *out, int maxlen)
{
    return dn_to_str(desc, dn, out, maxlen, 1);
}

/*
 * Extract a single component out of a distinguished name
 */
asn1_retCode cert_dnattr2str(ASN1_TYPE desc, ASN1_TYPE dn,
                             const char *oid, char *out, int maxlen)
{
    asn1_retCode result;
    int i, j, rdns, avas;
    char namebuf[ASN1_MAX_NAME_SIZE];
    char val[MAX_RDN_TEXT_SIZE];
    int val_len;

    assert(out);

    result = asn1_number_of_elements(dn, "rdnSequence", &rdns);
    if (result != ASN1_SUCCESS)
        return result;

    /*
     * Iterate over each RDN (Relative Distinguished Name)
     */
    for (i = rdns; i > 0; i--) {
        /*
         * Iterate over each AVA (Attribute Value Assertion) for
         * this RDN. Normally there is just one.
         */
        snprintf(namebuf, sizeof(namebuf), "rdnSequence.?%d", i);

        result = asn1_number_of_elements(dn, namebuf, &avas);
        if (result != ASN1_SUCCESS)
            return result;

        for (j = avas; j > 0; j--) {
            snprintf(namebuf, sizeof(namebuf), "rdnSequence.?%d.?%d.type", i,
                     j);

            val_len = sizeof(val) - 1;
            result = asn1_read_value(dn, namebuf, val, &val_len);
            if (result != ASN1_SUCCESS)
                return result;

            /* The component we're looking for? */
            if (!oideq(val, oid))
                continue;

            snprintf(namebuf, sizeof(namebuf), "rdnSequence.?%d.?%d.value", i,
                     j);

            val_len = sizeof(val) - 1;
            result = asn1_read_value(dn, namebuf, val, &val_len);
            if (result != ASN1_SUCCESS)
                return result;

            result = der_to_str(val, val_len, out, maxlen);
            if (result != ASN1_SUCCESS)
                return result;

            return ASN1_SUCCESS;
        }
    }

    return ASN1_ELEMENT_NOT_FOUND;
}
