/*
 * Copyright 2008-2014 Pierre Ossman <ossman@cendio.se> for Cendio AB
 * Copyright 2009-2014 Karl Mikaelsson <derfian@cendio.se> for Cendio AB
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include "cert.h"


/*
 * Extract subject name. Output in RFC 4514 form.
 */
static int tl_certtool_subject(struct certificate *cert)
{
    int ret;
    char dn[MAX_DN_TEXT_SIZE];

    ret = cert_subject(cert, dn, sizeof(dn));

    if (ret) {
        fprintf(stderr,
                "ERROR: Could not read subject from certificate (error code %d)\n",
                ret);
        return ret;
    }

    printf("%s\n", dn);
    return 0;
}

/*
 * Extract subject name. Output in one of many conflicting Novell
 * forms.
 */
static int tl_certtool_subject_dotted_novell(struct certificate *cert)
{
    int ret;
    char dn[MAX_DN_TEXT_SIZE];

    ret = cert_subject_dotted_novell(cert, dn, sizeof(dn));

    if (ret) {
        fprintf(stderr,
                "ERROR: Could not read Novell subject from certificate (error code %d)\n",
                ret);
        return ret;
    }

    printf("%s\n", dn);
    return 0;
}

/*
 * Extract certificate issuer. Output in RFC 4514 form.
 */
static int tl_certtool_issuer(struct certificate *cert)
{
    int ret;
    char dn[MAX_DN_TEXT_SIZE];

    ret = cert_issuer(cert, dn, sizeof(dn));

    if (ret) {
        fprintf(stderr,
                "ERROR: Could not read issuer from certificate (error code %d)\n",
                ret);
        return ret;
    }

    printf("%s\n", dn);
    return 0;
}

/*
 * Determines if the certificate is for a CA.
 */
static int tl_certtool_is_ca(struct certificate *cert)
{
    int ret;

    ret = cert_is_ca(cert);
    if (ret < 0) {
        fprintf(stderr,
                "ERROR: Could not determine if certificate is for a CA (error code %d)\n",
                ret);
        return ret;
    }

    printf("%d\n", ! !ret);

    return 0;
}

/*
 * Extract the public key.
 */
static int tl_certtool_pubkey(struct certificate *cert)
{
    char exp[17], mod[8193];
    int ret, i;
    size_t explen, modlen;

    modlen = sizeof(mod);
    explen = sizeof(exp);

    ret = cert_rsa_public_key_components(cert, exp, &explen, mod, &modlen);
    if (ret) {
        fprintf(stderr,
                "ERROR: Couldn't get data from certificate (not an RSA key? - error code is %d)\n",
                ret);
        return ret;
    }

    printf("rsa");

    printf(" ");

    for (i = 0; i < explen; i++)
        printf("%02x", (unsigned int)(unsigned char)exp[i]);

    printf(" ");

    for (i = 0; i < modlen; i++)
        printf("%02x", (unsigned int)(unsigned char)mod[i]);

    printf("\n");

    return 0;
}

/*
 * Extract serial number from certificate. Output in hexadecimal form.
 */
static int tl_certtool_serial(struct certificate *cert)
{
    char serial[256];
    size_t len;
    int ret, i;

    len = sizeof(serial);

    ret = cert_serial(cert, serial, &len);
    if (ret != 0) {
        fprintf(stderr,
                "ERROR: Could not extract serial from certificate (error %d)\n",
                ret);
        return ret;
    }

    for (i = 0; i < len; i++)
        printf("%02x", (unsigned int)(unsigned char)serial[i]);
    printf("\n");

    return 0;
}

/*
 * Extract signature from certificate. Output in hexadecimal form.
 */
static int tl_certtool_signature(struct certificate *cert)
{
    char signature[4096];
    size_t len;

    enum pkix_algo algo;

    int ret, i;

    len = sizeof(signature);

    ret = cert_signature(cert, signature, &len);
    if (ret != 0) {
        fprintf(stderr,
                "ERROR: Could not extract signature from certificate (error %d)\n",
                ret);
        return ret;
    }

    ret = cert_signature_algo(cert, &algo);
    if (ret != 0) {
        fprintf(stderr,
                "ERROR: Could not extract signature algorithm from certificate (error %d)\n",
                ret);
        return ret;
    }

    switch (algo) {
    case PKIX_ALGO_RSA:
        printf("rsa");
        break;
    case PKIX_ALGO_RSA_MD2:
        printf("rsa-md2");
        break;
    case PKIX_ALGO_RSA_MD5:
        printf("rsa-md5");
        break;
    case PKIX_ALGO_RSA_SHA1:
        printf("rsa-sha1");
        break;
    case PKIX_ALGO_RSA_SHA256:
        printf("rsa-sha256");
        break;
    case PKIX_ALGO_RSA_SHA384:
        printf("rsa-sha384");
        break;
    case PKIX_ALGO_RSA_SHA512:
        printf("rsa-sha512");
        break;
    default:
        fprintf(stderr, "ERROR: Unknown signature algorithm (%d)\n", algo);
        return -EINVAL;
    }

    printf(" ");

    for (i = 0; i < len; i++)
        printf("%02x", (unsigned int)(unsigned char)signature[i]);
    printf("\n");

    return 0;
}

/*
 * Extract CRL locations from certificate. Output one URI on each
 * line. Returns an error code and writes error messages to stderr on
 * error.
 */
static int tl_certtool_crl(struct certificate *cert)
{
    const size_t NUM_CRLS = 16;
    int i;
    char *crls[16] = { NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL
    };

    i = cert_crl(cert, crls, NUM_CRLS);

    if (i != 0) {

        switch (-i) {
        case ENOENT:
            /* This is not an error, because CRL locations aren't
               mandatory. */
            fprintf(stderr,
                    "ERROR: This certificate does not contain any CRL distribution points.\n");
            return 0;
        case ENOMEM:
            fprintf(stderr, "ERROR: The CRL list was too large to handle.\n");
            break;
        default:
            fprintf(stderr,
                    "ERROR: Could not extract CRL distribution points from certificate (error %d)\n",
                    i);
            break;
        }
        return i;
    }


    for (i = 0; i < 16; i++) {
        if (crls[i]) {
            printf("%s\n", crls[i]);
            free(crls[i]);
        }
    }
    return 0;
}

/*
 * FIXME: Implement OCSP support, or try to offload it to existing
 * implementations. Not suitable for the scope of bug 3653 though.
 */

/*
static int
tl_certtool_ocsp(struct certificate *cert) {
    return 0;
};
*/

/*
 * Retrieve the certificate notValidBefore and notValidAfter dates.
 */
static int tl_certtool_expiry(struct certificate *cert)
{
    char notbefore[64];
    char notafter[64];
    int i;

    i = cert_valid_notbefore(cert, notbefore, 64);
    if (i != 0) {
        fprintf(stderr,
                "ERROR: Could not extract valid_notbefore from certificate (error %d)\n",
                i);
        return i;
    }

    i = cert_valid_notafter(cert, notafter, 64);
    if (i != 0) {
        fprintf(stderr,
                "ERROR: Could not extract valid_notafter date from certificate (error %d)\n",
                i);
        return i;
    }

    printf("%s %s\n", notbefore, notafter);
    return 0;
}

void usage()
{
    fprintf(stderr,
            "Extracts information from X.509 certificates in DER form passed on stdin.\n");
    fprintf(stderr, "Usage: tl-certtool <option> [option...]\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, " --pubkey        Extract certificate public key\n");
    fprintf(stderr,
            " --subject       Extract certificate subject name (RFC4514 form)\n");
    fprintf(stderr,
            " --novellsubject Extract certificate subject name (dotted Novell form)\n");
    fprintf(stderr, " --issuer        Extract the certificate issuer\n");
    fprintf(stderr,
            " --is-ca         Determine if the certificate is for a CA\n");
    fprintf(stderr, " --serial        Extract serial number of certificate\n");
    fprintf(stderr,
            " --expiry        Extract valid_notbefore and valid_notafter dates\n");
    fprintf(stderr, " --signature     Extract the certificate signature\n");
    fprintf(stderr, " --crl           Extract CRL distribution points\n");
    /* fprintf(stderr, " --ocsp          Extract OCSP locations\n"); */
}

struct funcmap {
    const char *arg;
    int (*func) (struct certificate * cert);
};

static struct funcmap funcmap[] = {
    {"--pubkey", tl_certtool_pubkey,},
    {"--subject", tl_certtool_subject,},
    {"--issuer", tl_certtool_issuer,},
    {"--is-ca", tl_certtool_is_ca,},
    {"--serial", tl_certtool_serial,},
    {"--expiry", tl_certtool_expiry,},
    {"--signature", tl_certtool_signature,},
    {"--crl", tl_certtool_crl,},
    {"--novellsubject", tl_certtool_subject_dotted_novell,},
//    { "--ocsp",         tl_certtool_ocsp, },
};

int main(int argc, char **argv)
{
    char der[65536];
    int der_length;
    int i, n, ret;
    struct certificate *cert;

    if (argc < 2) {
        usage();
        exit(1);
    };

    /* First just check the arguments */
    for (i = 1; i < argc; i++) {
        for (n = 0; n < sizeof(funcmap) / sizeof(*funcmap); n++) {
            if (!strcmp(argv[i], funcmap[n].arg))
                break;
        }
        if (n == sizeof(funcmap) / sizeof(*funcmap)) {
            usage();
            exit(1);
        }
    }

    der_length = fread(der, 1, sizeof(der), stdin);

    cert = cert_alloc();
    if (!cert) {
        fprintf(stderr, "ERROR: Failed to allocate certificate.\n");
        exit(1);
    }

    ret = cert_parse(der, der_length, cert);
    if (ret) {
        fprintf(stderr,
                "ERROR: Failed to parse certificate (error code %d)\n", ret);
        cert_free(cert);
        exit(1);
    }

    /* Call desired function(s) */
    for (i = 1; i < argc; i++) {
        for (n = 0; n < sizeof(funcmap) / sizeof(*funcmap); n++) {
            if (!strcmp(argv[i], funcmap[n].arg)) {
                ret = funcmap[n].func(cert);
                if (ret)
                    return ret;
                break;
            }
        }
    }

    cert_free(cert);
    return 0;
}
