/*
 * Copyright 2008-2014 Pierre Ossman <ossman@cendio.se> for Cendio AB
 * Copyright 2009-2014 Karl Mikaelsson <derfian@cendio.se> for Cendio AB
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include "cert.h"

/*
 * Extract all revoked certiciates. Output is serials in hexadecimal form.
 */
static int tl_crltool_list(struct crl *crl)
{
    int ret, n, i;
    size_t count;

    char serial[256];
    size_t len;

    ret = crl_count(crl);
    if (ret < 0) {
        fprintf(stderr,
                "ERROR: Could not get revocation list size (error code %d)\n",
                ret);
        return ret;
    }

    count = ret;

    for (n = 0; n < count; n++) {
        len = sizeof(serial);
        ret = crl_entry(crl, n, serial, &len);
        if (ret != 0) {
            fprintf(stderr,
                    "ERROR: Could not extract revocation list entry %d (error %d)\n",
                    n, ret);
            return ret;
        }

        for (i = 0; i < len; i++)
            printf("%02x", (unsigned int)(unsigned char)serial[i]);
        printf("\n");
    }

    return 0;
}

/*
 * Extract revocation list issuer. Output in RFC 4514 form.
 */
static int tl_crltool_issuer(struct crl *crl)
{
    int ret;
    char dn[MAX_DN_TEXT_SIZE];

    ret = crl_issuer(crl, dn, sizeof(dn));

    if (ret) {
        fprintf(stderr,
                "ERROR: Could not read issuer from revocation list (error code %d)\n",
                ret);
        return ret;
    }

    printf("%s\n", dn);
    return 0;
}

/*
 * Extract signature from revocation list. Output in hexadecimal form.
 */
static int tl_crltool_signature(struct crl *crl)
{
    char signature[4096];
    size_t len;

    enum pkix_algo algo;

    int ret, i;

    len = sizeof(signature);

    ret = crl_signature(crl, signature, &len);
    if (ret != 0) {
        fprintf(stderr,
                "ERROR: Could not extract signature from revocation list (error %d)\n",
                ret);
        return ret;
    }

    ret = crl_signature_algo(crl, &algo);
    if (ret != 0) {
        fprintf(stderr,
                "ERROR: Could not extract signature algorithm from revocation list (error %d)\n",
                ret);
        return ret;
    }

    switch (algo) {
    case PKIX_ALGO_RSA:
        printf("rsa");
        break;
    case PKIX_ALGO_RSA_MD2:
        printf("rsa-md2");
        break;
    case PKIX_ALGO_RSA_MD5:
        printf("rsa-md5");
        break;
    case PKIX_ALGO_RSA_SHA1:
        printf("rsa-sha1");
        break;
    case PKIX_ALGO_RSA_SHA256:
        printf("rsa-sha256");
        break;
    case PKIX_ALGO_RSA_SHA384:
        printf("rsa-sha384");
        break;
    case PKIX_ALGO_RSA_SHA512:
        printf("rsa-sha512");
        break;
    default:
        fprintf(stderr, "ERROR: Unknown signature algorithm (%d)\n", algo);
        return -EINVAL;
    }

    printf(" ");

    for (i = 0; i < len; i++)
        printf("%02x", (unsigned int)(unsigned char)signature[i]);
    printf("\n");

    return 0;
}

/*
 * Retrieve the revocation list thisUpdate and nextUpdate dates.
 */
static int tl_crltool_update(struct crl *crl)
{
    char thisupdate[64];
    char nextupdate[64];
    int i;

    i = crl_update_this(crl, thisupdate, 64);
    if (i != 0) {
        fprintf(stderr,
                "ERROR: Could not extract thisUpdate from revocation list (error %d)\n",
                i);
        return i;
    }

    i = crl_update_next(crl, nextupdate, 64);
    if (i != 0) {
        fprintf(stderr,
                "ERROR: Could not extract nextUpdate date from revocation list (error %d)\n",
                i);
        return i;
    }

    printf("%s %s\n", thisupdate, nextupdate);
    return 0;
}

void usage()
{
    fprintf(stderr,
            "Extracts information from X.509 certificate revocation lists in DER form passed on stdin.\n");
    fprintf(stderr, "Usage: tl-crltool <option> [option...]\n");
    fprintf(stderr, "Options:\n");

    fprintf(stderr,
            " --list        Extract all serials in the revocation list\n");
    fprintf(stderr, " --issuer      Extract the revocation list issuer\n");
    fprintf(stderr, " --update      Extract thisUpdate and nextUpdate dates\n");
    fprintf(stderr, " --signature   Extract the revocation list signature\n");
}

struct funcmap {
    const char *arg;
    int (*func) (struct crl * crl);
};

static struct funcmap funcmap[] = {
    {"--list", tl_crltool_list,},
    {"--issuer", tl_crltool_issuer,},
    {"--update", tl_crltool_update,},
    {"--signature", tl_crltool_signature,},
};

#define ALLOC_CHUNK 65536

int main(int argc, char **argv)
{
    char *der;
    int der_length, alloc_length;
    int i, n, ret;
    struct crl *crl;

    if (argc < 2) {
        usage();
        exit(1);
    };

    /* First just check the arguments */
    for (i = 1; i < argc; i++) {
        for (n = 0; n < sizeof(funcmap) / sizeof(*funcmap); n++) {
            if (!strcmp(argv[i], funcmap[n].arg))
                break;
        }
        if (n == sizeof(funcmap) / sizeof(*funcmap)) {
            usage();
            exit(1);
        }
    }

    der = NULL;
    der_length = 0;
    alloc_length = 0;
    while (der_length == alloc_length) {
        alloc_length += ALLOC_CHUNK;
        der = realloc(der, alloc_length);

        der_length += fread(der + alloc_length - ALLOC_CHUNK, 1,
                            ALLOC_CHUNK, stdin);
    }

    crl = crl_alloc();
    if (!crl) {
        fprintf(stderr, "ERROR: Failed to allocate crl.\n");
        exit(1);
    }

    ret = crl_parse(der, der_length, crl);
    if (ret) {
        fprintf(stderr,
                "ERROR: Failed to parse revocation list (error code %d)\n",
                ret);
        crl_free(crl);
        exit(1);
    }

    /* Call desired function(s) */
    for (i = 1; i < argc; i++) {
        for (n = 0; n < sizeof(funcmap) / sizeof(*funcmap); n++) {
            if (!strcmp(argv[i], funcmap[n].arg)) {
                ret = funcmap[n].func(crl);
                if (ret)
                    return ret;
                break;
            }
        }
    }

    crl_free(crl);

    return 0;
}
